import potentiostat

model = 'emstatpico'
path = '.'
folder = 'data'
potentiostat.Setup(model, path, folder)

#cv = potentiostat.CV(sr=0.5)
#cv.bipot()
#cv.run()

lsv = potentiostat.LSV(sr=0.5)
lsv.bipot()
lsv.run()

#ca = potentiostat.CA()
#ca.bipot()
#ca.run()

#ocp = potentiostat.OCP()
#ocp.run()



import numpy as np
import softpotato as sp
import pypotato.mscript as mscript

class Test:
    '''
    '''
    def __init__(self):
        print('Test from save_data module')



class Save:
    '''
    '''
    def __init__(self, data, fileName, header, model, technique, bpot=0):
        self.fileName = fileName
        self.data_array = 0
        if technique == 'CV' or technique == 'LSV':
            header = header + '\nt/s, E/V, i/A\n' 
            self.data_array = CV(fileName, data, model, bpot).save()
        elif technique == 'IT' or technique == 'CA':
            header = header + '\nt/s, E/V, i/A\n'
            self.data_array = IT(fileName, data, model, bpot).save()
        elif technique == 'OCP':
            header = header + '\nt/s, E/V\n'
            self.data_array = OCP(fileName, data, model).save()
        np.savetxt(fileName, self.data_array, delimiter=',', header=header)


class CV:
    '''
    '''
    def __init__(self, fileName, data, model, bpot):
        self.fileName = fileName
        self.data = data
        self.model = model
        self.bpot = bpot
        data_array = 0

    def save(self):
        if self.model == 'emstatpico':
            t = mscript.get_values_by_column(self.data,0)
            E = mscript.get_values_by_column(self.data,1)
            i = mscript.get_values_by_column(self.data,2)
            data_array = np.array([t,E,i]).T
            if self.bpot:
                i2 = mscript.get_values_by_column(self.data,3)
                data_array = np.array([t,E,i,i2]).T

        return data_array



class IT:
    '''
    '''
    def __init__(self, fileName, data, model, bpot):
        self.fileName = fileName
        self.data = data
        self.model = model
        self.bpot = bpot
        data_array = 0

    def save(self):
        if self.model == 'emstatpico':
            t = mscript.get_values_by_column(self.data,0)
            E = mscript.get_values_by_column(self.data,1)
            i = mscript.get_values_by_column(self.data,2)
            data_array = np.array([t,E,i]).T
            if self.bpot:
                i2 = mscript.get_values_by_column(self.data,3)
                data_array = np.array([t,E,i,i2]).T
        return data_array


class OCP:
    '''
    '''
    def __init__(self, fileName, data, model):
        self.fileName = fileName
        self.data = data
        self.model = model
        data_array = 0

    def save(self):
        if self.model == 'emstatpico':
            t = mscript.get_values_by_column(self.data,0)
            E = mscript.get_values_by_column(self.data,1)
            #i = mscript.get_values_by_column(self.data,2)
            data_array = np.array([t,E]).T
        return data_array




"""
PalmSens instrument module

This module implements the communication protocol. This consists of the
high-level read and write methods, methods to read the firmware version,
identify the device type, execute scripts on the device, etc.

The low-level (physical) communication interface,  is implemented in another
module, so that multiple physical interfaces (e.g. serial port, USB, etc.)
can be supported.

MethodSCRIPT specific methods, such as parsing and interpreting the measurement
data, is implemented in the mscript module.

-------------------------------------------------------------------------------
Copyright (c) 2021 PalmSens BV
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

   - Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   - Neither the name of PalmSens BV nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.
   - This license does not release you from any requirement to obtain separate 
	  licenses from 3rd party patent holders to use this software.
   - Use of the software either in source or binary form must be connected to, 
	  run on or loaded to an PalmSens BV component.

DISCLAIMER: THIS SOFTWARE IS PROVIDED BY PALMSENS "AS IS" AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

# Standard library imports
import logging
import time


LOG = logging.getLogger(__name__)


class DeviceType:
    UNKNOWN = 'unknown device'
    EMSTAT_PICO = 'EmStat Pico'
    EMSTAT4_HR = 'EmStat4 HR'
    EMSTAT4_LR = 'EmStat4 LR'
    MULTI_EMSTAT4_HR = 'MultiEmStat4 HR'
    MULTI_EMSTAT4_LR = 'MultiEmStat4 LR'
    EMSTAT_PICO_BOOTLOADER = 'EmStat Pico bootloader'


_FIRMWARE_VERSION_TO_DEVICE_TYPE_MAPPING = [
    ('espico', DeviceType.EMSTAT_PICO),
    ('es4_hr', DeviceType.EMSTAT4_HR),
    ('es4_lr', DeviceType.EMSTAT4_LR),
    ('mes4hr', DeviceType.MULTI_EMSTAT4_HR),
    ('mes4lr', DeviceType.MULTI_EMSTAT4_LR),
    ('espbl', DeviceType.EMSTAT_PICO_BOOTLOADER),
]


class CommunicationError(Exception):
    """Generic communication error class."""


class CommunicationTimeout(Exception):
    """Communication timeout.

    Note that a communication timeout does not have to be an error. If a long
    measurement is running, it is possible that a communication timeout occurs
    while waiting on the response. In that case, just keep trying to read and
    (optionally) handle a global timeout in the calling method.
    This exception could be avoided by increasing the timeout on the low-level
    (serial) interface. However, that could cause the application to block and
    become unresponsive. It's better to keep the low-level read timeouts low
    (< 1 s) and handle timeouts at the application level.
    """


class Instrument():
    """Communication interface for MethodSCRIPT instruments.

    This class contains high-level communication methods that are independent
    of the physical interface (e.g.: serial port, USB, Bluetooth, ...). The
    low-level communication should be provided by a communication object that
    is passed to the initializer.

    The low-level communication module should only implement the following
    two methods:
        - write(data: bytes)
        - readline() -> bytes
    """

    def __init__(self, comm):
        """Initialize the object.

        `comm` must be a communication object as described in the
        documentation of this class.
        """
        self.comm = comm
        self.firmware_version = None
        self.device_type = DeviceType.UNKNOWN

    def write(self, text: str):
        """Write to device."""
        # The text is encoded using ASCII encoding, since all MethodSCRIPT
        # commands are plain ASCII text. String literals (as used in the
        # `send_string` command) and comments *could* contain non-ASCII
        # characters, but this is not officially supported or recommended.
        # Although using another encoding here could be beneficial for some
        # users, it could lead to unexpected problems for other users.
        # Therefore, the ASCII encoding is chosen, which is always safe and
        # easy to use as long as the input MethodSCRIPT does not contain
        # non-ASCII characters.
        data = text.encode('ascii')
        LOG.debug('TX: %r', data)
        self.comm.write(data)

    def writelines(self, lines):
        """Write multiple lines to the device."""
        for line in lines:
            self.write(line)

    def readline(self) -> str:
        """Read one response line from the device."""
        # Read line using the raw serial interface.
        data = self.comm.readline()
        # If we received data (i.e., no timeout), log it for debugging.
        if data:
            LOG.debug('RX: %r', data)
        # Decode the received line. Note that only ASCII characters are
        # expected (see the comment in `write()`). To avoid exceptions
        # in case invalid data is received, invalid bytes will be replaced
        # by a replacement character.
        line = data.decode('ascii', errors='replace')
        if not line:
            raise CommunicationTimeout()
        if line[-1] != '\n':
            raise CommunicationError('No EOL character received.')
        return line

    def readlines_until_end(self):
        """Receive all lines until an empty line is received."""
        lines = []
        print('Reading')
        while True:
            #print('Reading')
            try:
                line = self.readline()
            except CommunicationTimeout:
                continue
            if line == '\n':
                break
            lines.append(line)
        print('Finished reading')
        return lines

    def _update_firmware_version_and_device_type(self, force=False):
        # First get the firmware version string from the device.
        if force or not self.firmware_version:
            self.write('t\n')
            line1 = self.readline()
            line2 = self.readline()
            if not (line1.startswith('t') and line2.endswith('*\n')):
                raise CommunicationError('Invalid response to firmware version request.')
            self.firmware_version = (line1 + line2).replace('\n', ' ')[1:-1]
        # Then derive the device type from the firmware version string.
        for device_id, device_type in _FIRMWARE_VERSION_TO_DEVICE_TYPE_MAPPING:
            if self.firmware_version.startswith(device_id):
                self.device_type = device_type
                break
            else:
                self.device_type = DeviceType.UNKNOWN

    def get_firmware_version(self, force=False):
        """Get the device firmware version.

        The result of this call is cached. If it is changed on the device, use
        `force=true` to force reading it from the device again.
        """
        self._update_firmware_version_and_device_type(force=force)
        return self.firmware_version

    def get_device_type(self, force=False):
        """Get the device type.

        The result of this call is cached. If it is changed on the device, use
        `force=true` to force reading it from the device again.
        """
        self._update_firmware_version_and_device_type(force=force)
        return self.device_type

    def get_mscript_version(self):
        self.write('v\n')
        response = self.readline()
        return int(response[1:-1])

    def get_serial_number(self):
        """Read the EmStat Pico serial number."""
        self.write('i\n')
        return self.readline()[1:-1]

    def get_register(self, register):
        """Get the value of a register."""
        self.write('G%02d\n' % register)
        return self.readline()[1:-1]

    def load_mscript_from_flash(self):
        """Load the MethodSCRIPT from flash to RAM."""
        self.write('Lmscr\n')
        self.readline()
        # TODO: check response!

    def run_mscript_from_flash(self):
        """Load the MethodSCRIPT from flash to RAM and execute it."""
        self.write('Lmscr\n')
        self.readline()
        # TODO: check response!
        self.write('r\n')

    def send_script(self, path):
        """Read a script from file and send it to the device.

        Note that the file should contain ASCII characters only. Other
        characters or encodings are not supported. The file may contain
        any common end-of-line style (e.g. Unix or Windows line endings).
        The lines written to the device will always use '\n' line endings
        (Linux format).
        """
        with open(path, 'rt', encoding='ascii') as file:
            lines = file.readlines()
        self.writelines(lines)

    def abort_and_sync(self):
        """Abort a possibly running script and wait for it to finish.

        This method tries to get the device in a known valid state by sending an
        abort command and checking the response. If a script was still running, it
        will wait for it to complete. Note that this could take long, depending on
        the measurement that was running.

        Note that it should normally not be necessary to call this method, but it
        could be useful in case the Python script was interrupted or the serial
        communication was lost during a measurement. In that case, when restarting
        the script, it would receive data from the previous measurement, which
        would cause communication issues.
        This method should recover from such situation and restore communication.
        """
        LOG.info('Aborting possible active scripts and syncing communication.')
        # Send new line character to flush possible command in command buffer.
        self.write('\n')
        # Send abort command.
        self.write('Z\n')
        # Wait for acknowledgment of abort command.
        while True:
            response = self.readline()
            if response.startswith('Z'):
                break
        # Check response.
        if response == 'Z!0006\n':
            LOG.info('No active scripts are currently running.')
            # Wait for > 50 ms after a failed command ('!' in response).
            time.sleep(0.1)
        if response == 'Z\n':
            LOG.info('Waiting for active script to finish...')
            self.readlines_until_end()
        LOG.info('Device is ready.')



import os
import numpy as np
import pypotato.load_data as load_data
import pypotato.save_data as save_data
import softpotato as sp
import pypotato.chi760e as chi760e
import pypotato.chi1205b as chi1205b
import pypotato.emstatpico as emstatpico

import pypotato.instrument as instrument
import pypotato.mscript as mscript
import pypotato.serial as serial

# Potentiostat models available: chi760e, chi1205b

# Global variables
folder_save = '.'
model_pstat = 'no pstat'
path_lib = '.'

class Test:
    '''
    '''
    def __init__(self):
        print('Test from potentiostat module')

class Setup:
    def __init__(self, model=0, path='.', folder='.', port=None, verbose=1):
        global folder_save
        folder_save = folder
        global model_pstat
        model_pstat = model
        global path_lib
        path_lib = path
        global port_
        port_ = port
        if verbose:
            self.info()

    def info(self):
        print('\n----------')
        print('Potentiostat model: ' + model_pstat)
        print('Potentiostat path: ' + path_lib)
        print('Save folder: ' + folder_save)
        print('----------\n')


class Technique:
    '''
    '''

    def __init__(self, text='', fileName='CV'):
        self.text = text # text to write as macro
        self.fileName = fileName
        self.technique = 'Technique'
        self.bpot = False

    def writeToFile(self):
        if model_pstat[0:3] == 'chi':
            file = open(folder_save + '/' + self.fileName + '.mcr', 'wb')
            file.write(self.text.encode('ascii'))
            file.close()
        elif model_pstat == 'emstatpico':
            file = open(folder_save + '/' + self.fileName + '.mscr', 'wb')
            file.write(self.text.encode('ascii'))
            file.close()

    def run(self):
        if model_pstat[0:3] == 'chi':
            self.message()
            # Write macro:
            self.writeToFile()
            # Run command:
            command = path_lib #+ '/chi760e.exe'
            param = ' /runmacro:\"' + folder_save + '/' + self.fileName + '.mcr\"'
            #print(param)
            os.system(command + param)
            self.message(start=False)
            self.plot()
        elif model_pstat == 'emstatpico':
            self.message()
            self.writeToFile()
            if port_ is None:
                self.port = serial.auto_detect_port()
            with serial.Serial(self.port,1) as comm:
                dev = instrument.Instrument(comm)
                dev.send_script(folder_save + '/' + self.fileName + '.mscr')
                result = dev.readlines_until_end()
                #print(result)
            self.data = mscript.parse_result_lines(result)
            fileName = folder_save + '/' + self.fileName + '.txt'
            save = save_data.Save(self.data, fileName, self.header, model_pstat, 
                           self.technique, bpot=self.bpot)
            self.message(start=False)
            self.plot()
        else:
            print('\nNo potentiostat selected. Aborting.')

    def plot(self):
        figNum = np.random.randint(1000)
        #print(figNum)
        if self.technique == 'CV':
            cv = load_data.CV(self.fileName+'.txt', folder_save, model_pstat)
            sp.plotting.plot(cv.E, cv.i, show=False, fig=figNum,
                             fileName=folder_save + '/' + self.fileName)
        elif self.technique == 'LSV':
            lsv = load_data.LSV(self.fileName+'.txt', folder_save, model_pstat)
            sp.plotting.plot(lsv.E, lsv.i, show=False, fig=figNum,
                             fileName=folder_save + '/' + self.fileName)
        elif self.technique == 'CA':
            ca = load_data.CA(self.fileName+'.txt', folder_save, model_pstat)
            sp.plotting.plot(ca.t, ca.i, show=False, fig=figNum,
                             xlab='$t$ / s', ylab='$i$ / A',
                             fileName=folder_save + '/' + self.fileName)
        elif self.technique == 'OCP':
            ocp = load_data.OCP(self.fileName+'.txt', folder_save, model_pstat)
            sp.plotting.plot(ocp.t, ocp.E, show=False, fig=figNum,
                             xlab='$t$ / s', ylab='$E$ / V',
                             fileName=folder_save + '/' + self.fileName)
         


    def message(self, start=True):
        if start:
            print('----------\nStarting ' + self.technique)
            if self.bpot:
                print('Running in bipotentiostat mode')
        else:
            print(self.technique + ' finished\n----------\n')

    def bipot(self, E=-0.2, sens=1e-6):
        if self.technique != 'OCP' and self.technique != 'EIS':
            if model_pstat == 'chi760e':
                self.tech.bipot(E, sens)
                self.text = self.tech.text
                self.bpot = True
            elif model_pstat == 'emstatpico':
                self.tech.bipot(E, sens)
                self.text = self.tech.text
                self.bpot = True
        else:
            print(self.technique + ' does not have bipotentiostat mode')
     

class CV(Technique):
    '''
        resistance = ## in ohms in case manual IR compensation is required
        this option is not tested yet and it is only implemented in CV,
        if it works it will be implemented in other techniques
    '''
    def __init__(self, Eini=-0.2, Ev1=0.2, Ev2=-0.2, Efin=-0.2, sr=0.1,
                 dE=0.001, nSweeps=2, sens=1e-6,
                 fileName='CV', header='CV', resistance=0):
        self.header = header
        if model_pstat == 'chi760e':
            self.tech = chi760e.CV(Eini, Ev1, Ev2, Efin, sr, dE, nSweeps, sens,
                               folder_save, fileName, header, path_lib, qt=2, 
                               resistance=resistance)
            Technique.__init__(self, text=self.tech.text, fileName=fileName)
            self.technique = 'CV'
            print('CV')
        elif model_pstat == 'chi1205b':
            self.tech = chi1205b.CV(Eini, Ev1, Ev2, Efin, sr, dE, nSweeps, sens,
                               folder_save, fileName, header, path_lib, qt=2, 
                               resistance=resistance)
            Technique.__init__(self, text=self.tech.text, fileName=fileName)
            self.technique = 'CV'
            print('CV')
        elif model_pstat == 'emstatpico':
            self.tech = emstatpico.CV(Eini, Ev1, Ev2, Efin, sr, dE, nSweeps, sens, 
                                folder_save, fileName, header, path_lib='',
                                qt=0, resistance=0)
            Technique.__init__(self, text=self.tech.text, fileName=fileName)
            self.technique = 'CV'
            print('CV')


class NPV(Technique):
    '''
    '''
    def __init__(self, Eini=0.5, Efin=-0.5, dE=0.01, tsample=0.1, twidth=0.05, tperiod=10, sens=1e-6,
                 fileName='NPV', header='NPV performed with CHI760'):
        if model_pstat == 'chi760e':
            self.tech = chi760e.NPV(Eini, Efin, dE, tsample, twidth, tperiod, sens,
                         folder_save, fileName, header, path_lib, qt=0)
            Technique.__init__(self, text=self.tech.text, fileName=fileName)
            self.technique = 'NPV'
            print('NPV')

    


class LSV(Technique):
    '''
    '''
    def __init__(self, Eini=-0.2, Efin=0.2, sr=0.1, dE=0.001, sens=1e-6,
                 fileName='LSV', header='LSV'):
        self.header = header
        if model_pstat == 'chi760e':
            self.tech = chi760e.LSV(Eini, Efin, sr, dE, sens, folder_save, fileName, 
                                header, path_lib, qt=2)
            Technique.__init__(self, text=self.tech.text, fileName=fileName)
            self.technique = 'LSV'  
        elif model_pstat == 'chi1205b':
            self.tech = chi1205b.LSV(Eini, Efin, sr, dE, sens, folder_save, fileName, 
                                header, path_lib, qt=2)
            Technique.__init__(self, text=self.tech.text, fileName=fileName)
            self.technique = 'LSV'  
        elif model_pstat == 'emstatpico':
            self.tech = emstatpico.LSV(Eini, Efin, sr, dE, sens, folder_save, fileName, 
                                header, path_lib, qt=2)
            Technique.__init__(self, text=self.tech.text, fileName=fileName)
            self.technique = 'LSV'  



class CA(Technique):
    '''
    '''
    def __init__(self, Estep=0.2, dt=0.001, ttot=2, sens=1e-6,
                 fileName='CA', header='CA'):
        self.header = header
        if model_pstat == 'chi760e':
            self.tech = chi760e.CA(Estep, dt, ttot, sens, folder_save, fileName,
                               header, path_lib, qt=2)
            Technique.__init__(self, text=self.tech.text, fileName=fileName)
            self.technique = 'CA'
            print('CA')
        elif model_pstat == 'chi1205b':
            self.tech = chi1205b.CA(Estep, dt, ttot, sens, folder_save, fileName,
                               header, path_lib, qt=2)
            Technique.__init__(self, text=self.tech.text, fileName=fileName)
            self.technique = 'CA'
        elif model_pstat == 'emstatpico':
            self.tech = emstatpico.CA(Estep, dt, ttot, sens, folder_save, fileName,
                               header, path_lib, qt=2)
            Technique.__init__(self, text=self.tech.text, fileName=fileName)
            self.technique = 'CA'




class OCP(Technique):
    '''
    '''
    def __init__(self, ttot=2, dt=0.01, fileName='OCP', header='OCP'):
        self.header = header
        if model_pstat == 'chi760e':
            self.tech = chi760e.OCP(ttot, dt, folder_save, fileName, header, path_lib)
            Technique.__init__(self, text=self.tech.text, fileName=fileName)
            self.technique = 'OCP'
            print('OCP')
        elif model_pstat == 'chi1205b':
            self.tech = chi1205b.OCP(ttot, dt, folder_save, fileName, header, path_lib)
            Technique.__init__(self, text=self.tech.text, fileName=fileName)
            self.technique = 'OCP'
            print('OCP')
        elif model_pstat == 'emstatpico':
            self.tech = emstatpico.OCP(ttot, dt, folder_save, fileName, header, path_lib)
            Technique.__init__(self, text=self.tech.text, fileName=fileName)
            self.technique = 'OCP'
            print('OCP')



class EIS(Technique):
    '''
    '''
    def __init__(self, Eini=0, low_freq=1, high_freq=1000, amplitude=0.01, 
                 sens=1e-6, qt=0, fileName='EIS', header='EIS'):
        self.header = header
        if model_pstat == 'chi760e':
            self.tech = chi760e.EIS(Eini, low_freq, high_freq, amplitude, sens, qt, 
                                folder_save, fileName, header, path_lib)
            Technique.__init__(self, text=self.tech.text, fileName=fileName)
            self.technique = 'EIS'



if __name__ == '__main__':
    sens = 1e-8
    sr = [0.1, 0.2, 0.5]
    folder = 'C:/Users/oliverrz/Desktop/Oliver/Data/220113_PythonMacros'



class Test:
    '''
    '''
    def __init__(self):
        print('Test from Emstat Pico translator')


class CV:
    '''
    '''
    def __init__(self, Eini, Ev1, Ev2, Efin, sr, dE, nSweeps, sens,
                 folder, fileName, header, path_lib=None, qt=2, resistance=0):
        '''
            Potential based variables need to be changed to mV int(Eini*100).
            For some reason Pico does not accept not having prefix 
        '''
        self.Eini = int(Eini*1000)
        self.Ev1 = int(Ev1*1000)
        self.Ev2 = int(Ev2*1000)
        self.Efin = int(Efin*1000)
        self.sr = int(sr*1000)
        self.dE = int(dE*1000)
        self.nSweeps = nSweeps
        self.text = ''
        self.ini = 'e\nvar c\nvar p\nvar a\n'
        self.pre_body = 'set_pgstat_mode 4\nset_autoranging ba 100n 5m' +\
                        '\nset_e '+ str(self.Eini) + 'm\ncell_on\nwait 2\ntimer_start'
        self.body = '\nmeas_loop_cv p c ' + str(self.Eini) + 'm ' +\
                    str(self.Ev1) + 'm ' +\
                    str(self.Ev2) + 'm ' + str(self.dE) + 'm ' + str(self.sr) +\
                    'm nscans(' + str(self.nSweeps-1) + ')\n\tpck_start\n\ttimer_get a' +\
                    '\n\tpck_add a\n\tpck_add p\n\tpck_add c\n\tpck_end\nendloop\n' + \
                    'on_finished:\ncell_off\n\n'
        self.text = self.ini + self.pre_body + self.body


    def bipot(self, E, sens):
        E = int(E*1000)
        self.pre_body = 'var b\nset_pgstat_chan 1' +\
                        '\nset_pgstat_mode 5' +\
                        '\nset_poly_we_mode 0' +\
                        '\nset_e '+ str(E) + 'm\nset_autoranging ba 100n 5m' +\
                        '\nset_pgstat_chan 0\nset_pgstat_mode 2' +\
                        '\nset_autoranging ba 100n 5m\nset_e ' + str(self.Eini) +\
                        'm\ntimer_start\ncell_on'
        self.body = '\nmeas_loop_cv p c ' + str(self.Eini) + 'm ' +\
                    str(self.Ev1) + 'm ' +\
                    str(self.Ev2) + 'm ' + str(self.dE) + 'm ' + str(self.sr) +\
                    'm nscans(' + str(self.nSweeps) + ') poly_we(1 b)\n\t' +\
                    'pck_start\n\ttimer_get a' +\
                    '\n\tpck_add a\n\tpck_add p\n\tpck_add c\n\tpck_add b\n\t' +\
                    'pck_end\nendloop\non_finished:\ncell_off\n\n'
        self.text = self.ini + self.pre_body + self.body
        #print(self.text)

        


class CA:
    '''
    '''
    def __init__(self, Estep, dt, ttot, sens, folder, fileName, header,
                 path_lib=None, qt=2):
        '''
        '''
        self.Estep = int(Estep*1000)
        self.dt = int(dt*1000)
        self.ttot = int(ttot*1000)
        self.text = ''
        self.ini = 'e\nvar p\nvar c\nvar a\n'
        self.pre_body = 'set_pgstat_mode 3\nset_autoranging ba 100n 5m' +\
                        '\nset_e ' + str(Estep) + 'm\ncell_on\ntimer_start'
        self.body = '\nmeas_loop_ca p c ' + str(Estep) + 'm ' + str(dt) +\
                    'm ' + str(ttot) + 'm\n\tpck_start\n\ttimer_get a\n\t' +\
                    'pck_add a\n\t' +\
                    'pck_add p\n\tpck_add c\n\tpck_end\n\tendloop' +\
                    '\non_finished:\ncell_off\n\n'
        self.text = self.ini + self.pre_body + self.body

    def bipot(self, E, sens):
        E = int(E*1000)
        self.pre_body = 'var b\nset_pgstat_chan 1' +\
                        '\nset_pgstat_mode 5' +\
                        '\nset_poly_we_mode 0' +\
                        '\nset_e '+ str(E) + 'm\nset_autoranging ba 100n 5m' +\
                        '\nset_pgstat_chan 0\nset_pgstat_mode 2' +\
                        '\nset_autoranging ba 100n 5m\nset_e ' + str(self.Estep) +\
                        'm\ntimer_start\ncell_on'
        self.body = '\nmeas_loop_ca p c ' + str(self.Estep) + 'm ' +\
                    str(self.dt) + 'm ' +\
                    str(self.ttot) + 'm poly_we(1 b)\n\t' +\
                    'pck_start\n\ttimer_get a' +\
                    '\n\tpck_add a\n\tpck_add p\n\tpck_add c\n\tpck_add b\n\t' +\
                    'pck_end\nendloop\non_finished:\ncell_off\n\n'
        self.text = self.ini + self.pre_body + self.body

        pass


class LSV:
    '''
    '''
    def __init__(self, Eini, Efin, sr, dE, sens, folder, fileName, header, 
                 path_lib=None, qt=2):
        self.Eini = int(Eini*1000)
        self.Efin = int(Efin*1000)
        self.sr = int(sr*1000)
        self.dE = int(dE*1000)
        self.text = ''
        self.ini = 'e\nvar c\nvar p\nvar a\n'
        self.pre_body = 'set_pgstat_mode 4\nset_autoranging ba 100n 5m' +\
                        '\nset_e '+ str(Eini) + 'm\ncell_on\ntimer_start'
        self.body = '\nmeas_loop_lsv p c ' + str(Eini) + 'm ' + str(Efin) + 'm ' +\
                    str(dE) + 'm ' + str(sr) +\
                    'm\n\tpck_start\n\ttimer_get a' +\
                    '\n\tpck_add a\n\tpck_add p\n\tpck_add c\n\tpck_end\nendloop\n' + \
                    'on_finished:\ncell_off\n\n'
        self.text = self.ini + self.pre_body + self.body

    def bipot(self, E, sens):
        E = int(E*1000)
        self.pre_body = 'var b\nset_pgstat_chan 1' +\
                        '\nset_pgstat_mode 5' +\
                        '\nset_poly_we_mode 0' +\
                        '\nset_e '+ str(E) + 'm\nset_autoranging ba 100n 5m' +\
                        '\nset_pgstat_chan 0\nset_pgstat_mode 2' +\
                        '\nset_autoranging ba 100n 5m\nset_e ' + str(self.Eini) +\
                        'm\ntimer_start\ncell_on'
        self.body = '\nmeas_loop_lsv p c ' + str(self.Eini) + 'm ' +\
                    str(self.Efin) + 'm ' +\
                    str(self.dE) + 'm ' + str(self.sr) +\
                    'm poly_we(1 b)\n\t' +\
                    'pck_start\n\ttimer_get a' +\
                    '\n\tpck_add a\n\tpck_add p\n\tpck_add c\n\tpck_add b\n\t' +\
                    'pck_end\nendloop\non_finished:\ncell_off\n\n'
        self.text = self.ini + self.pre_body + self.body
        #print(self.text)


class OCP:
    '''
    '''
    def __init__(self, ttot, dt, folder, fileName, header, path_lib=None):
        dt = int(dt*1000)
        ttot = int(ttot*1000)
        self.text = ''
        self.ini = 'e\nvar p\nvar a\n'
        self.pre_body = 'set_pgstat_mode 4\ntimer_start\n'
        self.body = 'meas_loop_ocp p ' + str(dt) + 'm ' + str(ttot) + 'm '+\
                    '\n\tpck_start\n\ttimer_get a\n\tpck_add a\n\tpck_add p' +\
                    '\n\tpck_end\nendloop\non_finished:\ncell_off\n\n'
        self.text = self.ini + self.pre_body + self.body
        

        #'\n\tpck_start\n\ttimer_get a\n\tpck_add p\n\tpck_add a' +\



"""
PalmSens Serial Port (UART) interface

This module implements the serial interface to the PalmSens instrument.

This module uses the "pyserial" module, which must be installed before running
this code. See https://pypi.org/project/pyserial/ for more information.

-------------------------------------------------------------------------------
Copyright (c) 2021 PalmSens BV
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

   - Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   - Neither the name of PalmSens BV nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.
   - This license does not release you from any requirement to obtain separate 
	  licenses from 3rd party patent holders to use this software.
   - Use of the software either in source or binary form must be connected to, 
	  run on or loaded to an PalmSens BV component.

DISCLAIMER: THIS SOFTWARE IS PROVIDED BY PALMSENS "AS IS" AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

# Standard library imports
import logging

# Third-party imports
import serial
import serial.tools.list_ports


LOG = logging.getLogger(__name__)


def _is_mscript_device(port):
    """Check if the specified port is a known MethodSCRIPT device."""
    # NOTES:
    # - Since the EmStat Pico uses a generic FTDI USB-to-Serial chip,
    #   it is identified by Windows as "USB Serial Port". This is the
    #   text to look for when using the auto detection feature. Note
    #   that an EmStat Pico or Sensit BT cannot be auto-detected if
    #   there are also other devices connected that use this name.
    # - An EmStat4 device in bootloader mode would be identified as
    #   'EmStat4 Bootloader', but we only want to connect to devices
    #   that can run MethodSCRIPTs, so we do not include that here.
    return (port.description == 'EmStat4' or
            port.description.startswith('ESPicoDev') or
            port.description.startswith('SensitBT') or
            port.description.startswith('SensitSmart') or
            # ^ Above names are used in Linux
            # v Below names are used in Windows
            port.description.startswith('EmStat4 LR (COM') or
            port.description.startswith('EmStat4 HR (COM') or
            port.description.startswith('MultiEmStat4 LR (COM') or
            port.description.startswith('MultiEmStat4 HR (COM') or
            port.description.startswith('USB Serial Port'))


def auto_detect_port():
    """Auto detect serial communication port.

    This works by searching for an available port with the correct name.
    If exactly one port matches, this port will be returned. If there
    are either no or multiple matches, the auto detection fails and None
    is returned instead. In that case, the user must explicitly specify
    which port to connect to (or disconnect unneeded devices with the
    same port name).
    """
    LOG.info('Auto-detecting serial communication port.')
    # Get the available ports.
    ports = serial.tools.list_ports.comports(include_links=False)
    candidates = []
    for port in ports:
        LOG.debug('Found port: %s', port.description)
        if _is_mscript_device(port):
            candidates.append(port)

    if len(candidates) != 1:
        LOG.error('%d candidates found. Auto detect failed.', len(candidates))
        raise Exception('Auto detection of serial port failed.')

    LOG.info('Exactly one candidate found. Using %s.', port.device)
    return candidates[0].device


class Serial():
    """Serial communication interface for EmStat Pico."""

    def __init__(self, port, timeout):
        self.connection = serial.Serial(port=None, baudrate=230400, timeout=timeout)
        self.connection.port = port

    def __enter__(self):
        if not self.connection.is_open:
            self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def open(self):
        self.connection.open()

    def close(self):
        self.connection.close()

    def write(self, data: bytes):
        self.connection.write(data)

    def readline(self) -> bytes:
        return self.connection.readline()
